from functools import partial
from pathlib import Path

from pydantic import ValidationError
from pytest import importorskip, raises

from stjames import Atom, Molecule


def test_molecule_pbc() -> None:
    # this is a funny joke, because it makes manganese
    glomar_explorer = partial(Molecule, charge=0, multiplicity=2, atoms=[Atom(atomic_number=25, position=[0, 0, 0])])

    mol_nopbc = glomar_explorer()
    assert mol_nopbc.cell is None

    valid_input = ((1.0, 0.0, 0.0), (0.0, 1.0, 0.0), (0.0, 0.0, 1.0))
    mol = glomar_explorer(cell={"lattice_vectors": valid_input})
    assert mol.cell is not None
    assert mol.cell.lattice_vectors == valid_input

    mol2 = glomar_explorer(cell={"lattice_vectors": valid_input, "is_periodic": (True, False, True)})
    assert mol2.cell is not None
    assert mol2.cell.lattice_vectors == valid_input
    assert mol2.cell.is_periodic == (True, False, True)

    with raises(ValidationError):
        glomar_explorer(cell={"lattice_vectors": valid_input, "is_periodic": (False, False, False)})

    invalid_input = ((1.0, 0.0, 0.0), (0.0, 1.0, 0.0), (0.0, 0.0))
    with raises(ValidationError):
        glomar_explorer(cell={"lattice_vectors": invalid_input})

    invalid_input2 = ((1.0, 0.0, 0.0), (0.0, 1.0, 0.0), (0.0, 0.0, "invalid"))
    with raises(ValidationError):
        glomar_explorer(cell={"lattice_vectors": invalid_input2})

    invalid_input3 = ((1.0, 0.0, 0.0), (0.0, 1.0, 0.0), (0.0, 0.0, 1.0), (1.0, 1.0, 1.0))
    with raises(ValidationError):
        glomar_explorer(cell={"lattice_vectors": invalid_input3})


def test_xyz() -> None:
    mol = Molecule.from_xyz_lines(
        """3
H2O Molecule; smiles: O; nonsense: data: more nonsense
O        0.000000     0.000000     0.000000
H        0.000000    -0.757000     0.587000
H        0.000000     0.757000     0.587000
""".splitlines()
    )
    assert mol.smiles == "O"
    assert getattr(mol, "nonsense", None) is None

    mol = Molecule.from_xyz_lines(
        """3
name: H2O Molecule; smiles: O
O        0.000000     0.000000     0.000000
H        0.000000    -0.757000     0.587000
H        0.000000     0.757000     0.587000
""".splitlines()
    )

    assert getattr(mol, "name", None) is None

    mol = Molecule.from_xyz_lines(
        """3
cell: [(1, 2, 3), [4, 5, 6e-1], (7.0, 8.0, 9.0)]; smiles: O; is_periodic: [True, False, True]
O        0.000000     0.000000     0.000000
H        0.000000    -0.757000     0.587000
H        0.000000     0.757000     0.587000
""".splitlines()
    )

    assert mol.cell is not None
    assert mol.cell.lattice_vectors == ((1.0, 2.0, 3.0), (4.0, 5.0, 0.6), (7.0, 8.0, 9.0))
    assert mol.cell.is_periodic == (True, False, True)
    assert mol.smiles == "O"

    mol = Molecule.from_xyz_lines(
        """3
cell: [(a, b, c), [4, 5, 6e-1], (7.0, 8.0, 9.0)]; smiles: O; is_periodic: [True, False, True]
O        0.000000     0.000000     0.000000
H        0.000000    -0.757000     0.587000
H        0.000000     0.757000     0.587000
""".splitlines()
    )

    assert mol.cell is None
    assert mol.smiles == "O"
    assert getattr(mol.cell, "is_periodic", None) is None


# Only works if rdkit installed
def test_from_smiles() -> None:
    importorskip("rdkit")
    mol = Molecule.from_smiles("CCO")

    assert len(mol.atoms) == 9


SDF_TWO_MOLS = """\
benzene
     RDKit          3D

  6  6  0  0  0  0  0  0  0  0999 V2000
    1.2124    0.7000    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    1.2124   -0.7000    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    0.0000   -1.4000    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -1.2124   -0.7000    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
   -1.2124    0.7000    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
    0.0000    1.4000    0.0000 C   0  0  0  0  0  0  0  0  0  0  0  0
  1  2  2  0
  2  3  1  0
  3  4  2  0
  4  5  1  0
  5  6  2  0
  6  1  1  0
M  END
$$$$
water
     RDKit          3D

  3  2  0  0  0  0  0  0  0  0999 V2000
    0.0000    0.0000    0.0000 O   0  0  0  0  0  0  0  0  0  0  0  0
    0.9572    0.0000    0.0000 H   0  0  0  0  0  0  0  0  0  0  0  0
   -0.2399    0.9267    0.0000 H   0  0  0  0  0  0  0  0  0  0  0  0
  1  2  1  0
  1  3  1  0
M  END
$$$$
"""

MOL2_TWO_MOLS = """\
@<TRIPOS>MOLECULE
benzene
 6 6 0 0 0
SMALL
NO_CHARGES

@<TRIPOS>ATOM
      1 C1          1.2124    0.7000    0.0000 C.ar    1  LIG1        0.0000
      2 C2          1.2124   -0.7000    0.0000 C.ar    1  LIG1        0.0000
      3 C3          0.0000   -1.4000    0.0000 C.ar    1  LIG1        0.0000
      4 C4         -1.2124   -0.7000    0.0000 C.ar    1  LIG1        0.0000
      5 C5         -1.2124    0.7000    0.0000 C.ar    1  LIG1        0.0000
      6 C6          0.0000    1.4000    0.0000 C.ar    1  LIG1        0.0000
@<TRIPOS>BOND
     1     1     2   ar
     2     2     3   ar
     3     3     4   ar
     4     4     5   ar
     5     5     6   ar
     6     6     1   ar
@<TRIPOS>MOLECULE
water
 3 2 0 0 0
SMALL
NO_CHARGES

@<TRIPOS>ATOM
      1 O1          0.0000    0.0000    0.0000 O.3     1  LIG1        0.0000
      2 H1          0.9572    0.0000    0.0000 H       1  LIG1        0.0000
      3 H2         -0.2399    0.9267    0.0000 H       1  LIG1        0.0000
@<TRIPOS>BOND
     1     1     2   1
     2     1     3   1
"""


def test_molecules_from_sdf(tmp_path: Path) -> None:
    importorskip("rdkit")
    sdf_file = tmp_path / "test.sdf"
    sdf_file.write_text(SDF_TWO_MOLS)

    mols = Molecule.molecules_from_sdf(sdf_file)
    assert len(mols) == 2
    assert all(len(m.atoms) > 0 for m in mols)


def test_molecules_from_mol2(tmp_path: Path) -> None:
    importorskip("rdkit")
    mol2_file = tmp_path / "test.mol2"
    mol2_file.write_text(MOL2_TWO_MOLS)

    mols = Molecule.molecules_from_mol2(mol2_file)
    assert len(mols) == 2
    assert all(len(m.atoms) > 0 for m in mols)


def test_from_file_sdf(tmp_path: Path) -> None:
    importorskip("rdkit")
    sdf_file = tmp_path / "test.sdf"
    sdf_file.write_text(SDF_TWO_MOLS)

    mol = Molecule.from_file(sdf_file)
    assert len(mol.atoms) > 0


def test_from_file_mol2(tmp_path: Path) -> None:
    importorskip("rdkit")
    mol2_file = tmp_path / "test.mol2"
    mol2_file.write_text(MOL2_TWO_MOLS)

    mol = Molecule.from_file(mol2_file)
    assert len(mol.atoms) > 0



"""Tautomer prediction workflow."""

from typing import Annotated

from pydantic import AfterValidator

from ..base import Base, round_float, round_optional_float
from ..conformers import ConformerGenSettingsUnion, ETKDGSettings
from ..mode import Mode
from ..settings import Settings
from .multistage_opt import MultiStageOptSettings
from .workflow import DBCalculation, MoleculeWorkflow


class Tautomer(Base):
    """
    A tautomer.

    :param energy: energy of the tautomer
    :param weight: statistical weight of the tautomer
    :param predicted_relative_energy: relative energy of the tautomer
    :param structures: UUIDs of the structures
    """

    energy: Annotated[float, AfterValidator(round_float(6))]
    weight: Annotated[float | None, AfterValidator(round_optional_float(6))] = None
    predicted_relative_energy: Annotated[float | None, AfterValidator(round_optional_float(6))] = None

    structures: list[DBCalculation] = []


class TautomerWorkflow(MoleculeWorkflow):
    """
    A workflow to calculate tautomers.

    Inherited:
    :param initial_molecule: Molecule of interest
    :param mode: Mode for workflow (deprecated)

    New:
    :param conf_gen_settings: conformer-search settings, defaults to old "careful"
    :param multistage_opt_settings: optimization settings

    Results:
    :param tautomers: resulting Tautomers
    """

    conf_gen_settings: ConformerGenSettingsUnion = ETKDGSettings(
        num_initial_confs=250,
        max_confs=20,
        max_mmff_energy=15,
    )

    multistage_opt_settings: MultiStageOptSettings = MultiStageOptSettings(
        mode=Mode.MANUAL,
        optimization_settings=[Settings(method="aimnet2_wb97md3", tasks=["optimize"])],
        singlepoint_settings=Settings(method="aimnet2_wb97md3", tasks=["energy"], solvent_settings={"solvent": "water", "model": "cpcmx"}),
    )

    mode: Mode = Mode.CAREFUL  # deprecated
    tautomers: list[Tautomer] = []



"""Engine/method/correction/task compatibility tables.

Used by ``Settings`` validators to catch invalid combinations at construction time.
"""

from itertools import product

from .correction import Correction
from .engine import Engine
from .method import Method
from .solvent import SolventModel
from .task import Task

ENGINE_METHODS: dict[Engine, frozenset[Method]] = {
    Engine.AIMNET2: frozenset(
        {
            Method.AIMNET2_WB97MD3,
        }
    ),
    Engine.EGRET: frozenset(
        {
            Method.EGRET_1,
            Method.EGRET_1E,
            Method.EGRET_1T,
        }
    ),
    Engine.OMOL25: frozenset(
        {
            Method.OMOL25_CONSERVING_S,
            Method.UMA_S_OMOL,
            Method.UMA_S_1_2_OMOL,
            Method.UMA_M_OMOL,
            Method.UMA_S_OMAT,
            Method.UMA_S_1_2_OMAT,
            Method.UMA_M_OMAT,
            Method.UMA_S_OMC,
            Method.UMA_S_1_2_OMC,
            Method.UMA_M_OMC,
        }
    ),
    Engine.ORB: frozenset(
        {
            Method.ORB_V3_CONSERVATIVE_INF_OMAT,
            Method.ORB_V3_CONSERVATIVE_OMOL,
        }
    ),
    Engine.XTB: frozenset(
        {
            Method.GFN_FF,
            Method.GFN0_XTB,
            Method.GFN1_XTB,
            Method.GFN2_XTB,
            Method.G_XTB,
        }
    ),
    Engine.TBLITE: frozenset(
        {
            Method.GFN1_XTB,
            Method.GFN2_XTB,
        }
    ),
    Engine.OPENFF: frozenset(
        {
            Method.OFF_SAGE_2_0_0,
            Method.OFF_SAGE_2_2_1,
            Method.OFF_SAGE_2_3_0,
            Method.SMIRNOFF_2_0_0_AMBER_AM1BCC,
            Method.SMIRNOFF_2_2_1_AMBER_AM1BCC,
        }
    ),
    Engine.PSI4: frozenset(
        {
            Method.HARTREE_FOCK,
            Method.PBE,
            Method.R2SCAN,
            Method.TPSS,
            Method.M06L,
            Method.PBE0,
            Method.B3LYP,
            Method.TPSSH,
            Method.M06,
            Method.M062X,
            Method.CAMB3LYP,
            Method.WB97XV,
            Method.WB97XD3,
            Method.WB97MV,
            Method.WB97MD3BJ,
            Method.HF3C,
            Method.B973C,
            Method.B97D3BJ,
            Method.R2SCAN3C,
            Method.WB97X3C,
            Method.DSDBLYPD3BJ,
        }
    ),
    Engine.PYSCF: frozenset(
        {
            Method.HARTREE_FOCK,
            Method.BP86,
            Method.PBE,
            Method.R2SCAN,
            Method.TPSS,
            Method.M06L,
            Method.PBE0,
            Method.B3LYP,
            Method.TPSSH,
            Method.M06,
            Method.M062X,
            Method.CAMB3LYP,
            Method.WB97XV,
            Method.WB97XD3,
            Method.WB97MV,
            Method.WB97MD3BJ,
            Method.WB97X3C,
            Method.SKALA,
        }
    ),
    Engine.GPU4PYSCF: frozenset(
        {
            Method.HARTREE_FOCK,
            Method.BP86,
            Method.PBE,
            Method.R2SCAN,
            Method.TPSS,
            Method.M06L,
            Method.PBE0,
            Method.B3LYP,
            Method.TPSSH,
            Method.M06,
            Method.M062X,
            Method.CAMB3LYP,
            Method.WB97XV,
            Method.WB97XD3,
            Method.WB97MV,
            Method.WB97MD3BJ,
            Method.WB97X3C,
            Method.SKALA,
        }
    ),
}

METHOD_ENGINES: dict[Method, frozenset[Engine]] = {
    method: frozenset(engine for engine, methods in ENGINE_METHODS.items() if method in methods)
    for method in {m for methods in ENGINE_METHODS.values() for m in methods}
}

# Engine-level defaults; use get_supported_corrections() for method-specific overrides.
ENGINE_CORRECTIONS: dict[Engine, frozenset[Correction]] = {
    Engine.AIMNET2: frozenset(),
    Engine.EGRET: frozenset(),
    Engine.OMOL25: frozenset(),
    Engine.ORB: frozenset({Correction.D3}),
    Engine.XTB: frozenset(),
    Engine.TBLITE: frozenset(),
    Engine.OPENFF: frozenset(),
    Engine.PSI4: frozenset({Correction.D3BJ}),
    Engine.PYSCF: frozenset({Correction.D3, Correction.D3BJ, Correction.D4}),
    Engine.GPU4PYSCF: frozenset({Correction.D3, Correction.D3BJ, Correction.D4}),
}

# Per-(method, engine) overrides on top of ENGINE_CORRECTIONS.
# r2scan+psi4 and b3lyp allow D4; M06 family on psi4 only allows D3.
# Methods with dispersion baked in (WB97X-D3, WB97X-3C, WB97M-D3BJ, etc.) disallow further corrections.
_METHOD_ENGINE_CORRECTION_OVERRIDES: dict[tuple[Method, Engine], frozenset[Correction]] = {
    (Method.R2SCAN, Engine.PSI4): frozenset({Correction.D3BJ, Correction.D4}),
    (Method.B3LYP, Engine.PSI4): frozenset({Correction.D3BJ, Correction.D4}),
    (Method.B3LYP, Engine.PYSCF): frozenset({Correction.D3, Correction.D3BJ, Correction.D4}),
    (Method.B3LYP, Engine.GPU4PYSCF): frozenset({Correction.D3, Correction.D3BJ, Correction.D4}),
    (Method.M06, Engine.PSI4): frozenset({Correction.D3}),
    (Method.M06L, Engine.PSI4): frozenset({Correction.D3}),
    (Method.M062X, Engine.PSI4): frozenset({Correction.D3}),
    # Dispersion baked in — no additional corrections allowed
    **{
        (method, engine): frozenset()
        for method, engine in product(
            (Method.WB97XD3, Method.WB97X3C, Method.WB97MD3BJ, Method.WB97XV, Method.WB97MV, Method.DSDBLYPD3BJ, Method.B97D3BJ),
            (Engine.PSI4, Engine.PYSCF, Engine.GPU4PYSCF),
        )
    },
}

# Tasks unavailable per engine (periodic mode may disable additional tasks).
ENGINE_DISABLED_TASKS: dict[Engine, frozenset[Task]] = {
    Engine.AIMNET2: frozenset({Task.DIPOLE, Task.SPIN_DENSITY}),
    Engine.EGRET: frozenset({Task.CHARGE, Task.DIPOLE, Task.SPIN_DENSITY}),
    Engine.OMOL25: frozenset({Task.CHARGE, Task.DIPOLE, Task.SPIN_DENSITY}),
    Engine.ORB: frozenset({Task.CHARGE, Task.DIPOLE, Task.SPIN_DENSITY}),
    Engine.XTB: frozenset({Task.SPIN_DENSITY}),
    Engine.TBLITE: frozenset({Task.SPIN_DENSITY, Task.FREQUENCIES, Task.OPTIMIZE_TS}),
    Engine.OPENFF: frozenset(),
    Engine.PSI4: frozenset(),
    Engine.PYSCF: frozenset(),
    Engine.GPU4PYSCF: frozenset(),
}

# Engines that do not support open-shell (multiplicity > 1) calculations.
ENGINE_NO_OPEN_SHELL: frozenset[Engine] = frozenset(
    {
        Engine.AIMNET2,
        Engine.EGRET,
        Engine.ORB,
    }
)

# Engines that accept an explicit basis set.
ENGINE_SUPPORTS_BASIS_SET: frozenset[Engine] = frozenset(
    {
        Engine.PSI4,
        Engine.PYSCF,
        Engine.GPU4PYSCF,
    }
)


# Solvent models supported per engine. Engines not listed do not support solvent models.
ENGINE_SOLVENT_MODELS: dict[Engine, frozenset[SolventModel]] = {
    Engine.PSI4: frozenset({SolventModel.COSMO, SolventModel.CPCM, SolventModel.PCM}),
    Engine.PYSCF: frozenset({SolventModel.COSMO, SolventModel.CPCM, SolventModel.PCM}),
    Engine.GPU4PYSCF: frozenset({SolventModel.CPCM, SolventModel.PCM}),
    Engine.XTB: frozenset({SolventModel.ALPB, SolventModel.GBSA, SolventModel.CPCMX}),
    Engine.AIMNET2: frozenset({SolventModel.ALPB, SolventModel.CPCMX}),
    Engine.OMOL25: frozenset({SolventModel.ALPB, SolventModel.CPCMX}),
    Engine.ORB: frozenset({SolventModel.ALPB, SolventModel.CPCMX}),
    Engine.EGRET: frozenset({SolventModel.ALPB, SolventModel.CPCMX}),
}


def get_supported_corrections(method: Method, engine: Engine) -> frozenset[Correction]:
    """Return corrections supported for a given method/engine combination.

    Applies per-method overrides from ``_METHOD_ENGINE_CORRECTION_OVERRIDES``
    on top of ``ENGINE_CORRECTIONS``.

    :param method: computational method
    :param engine: compute engine
    :returns: supported corrections, empty if none
    """
    if (method, engine) in _METHOD_ENGINE_CORRECTION_OVERRIDES:
        return _METHOD_ENGINE_CORRECTION_OVERRIDES[(method, engine)]
    return ENGINE_CORRECTIONS.get(engine, frozenset())
